import com.haxepunk.HXP;
import com.haxepunk.Entity;
import com.haxepunk.utils.*;
import com.haxepunk.graphics.*;
import haxe.xml.Fast;

class Damage {
	public var type: String;
	public var amount: Float;
	public var duration: Float;
	public function new() {}
}

class Attack {
	public var start:    Float;		// Distance from origin to spawn particle
	public var range:    Float;		// Distance particle will move
	public var weight:   Int;		// Likelyhood of this attack being used (if multiple)
	public var damage:   Array<Damage>;		// Damage inflicted (types)
	public var speed :   Float;		// Particle speed
	public var cooldown: Int;		// frames until fire again
	public var image:    String;	// particle image
	public var rotated:  Bool;		// rotate particle with direction
	public var attached: Bool;		// Is the effect attached to the monster (melee)
	public function new() {}
}

class MonsterDef {
	public var name:    String;		// Monster type name
	public var desc:    String;		// Monster text
	public var speed:   Float;		// Movement speed
	public var health:  Float;		// Max health
	public var defence: Map<String, Float>;				// Defence multipliers
	public var image:   String;		// Graphic
	// AI
	public var threat:     Float;	// Threat generated to other greatures
	public var aggression: Float;	// How much we will attack
	public var range:      Float;	// Try to stay at this distance from target
	public var sense:      Float;	// sensor range
	// Attacks
	public var attacks = [];		// Array of attacks
}

enum MonsterState {
	SLEEP;
	IDLE;
	WANDER;
	ATTACK;
	FLEE;
}


class Monster extends Entity {
	private static var monsterList: Map<String, MonsterDef>; // Map of monster templates
	public static function loadMonsters(file:String) {
		if(monsterList==null) monsterList = new Map();
		var data:String = openfl.Assets.getText(file);
		if(data==null) return;
		try {
			var xml:Fast = new Fast( Xml.parse(data) );
			for(def in xml.node.monsters.nodes.monster) {
				if(!def.has.name) continue;
				var m:MonsterDef = new MonsterDef();
				// Data
				m.name    = def.att.name;
				m.image   = def.node.image.att.file;
				m.speed   = Std.parseFloat( def.node.stats.att.speed );
				m.health  = Std.parseFloat( def.node.stats.att.health );

				// AI
				var ai:Fast = def.node.ai;
				m.threat     = ai.has.threat?     Std.parseFloat( ai.att.threat ): 5;
				m.aggression = ai.has.aggression? Std.parseFloat( ai.att.aggression ): 0;
				m.range      = ai.has.range?      Std.parseFloat( ai.att.range ): 5;
				m.sense      = ai.has.sense?      Std.parseFloat( ai.att.sense ): 150;

				
				// Attacks
				for(atk in def.nodes.attack) {
					var a: Attack = new Attack();
					a.range    = atk.has.range?  Std.parseFloat( atk.att.range ):  0;
					a.weight   = atk.has.weight? Std.parseInt( atk.att.weight ): 1;
					a.cooldown = atk.has.speed?  Std.int(60 / Std.parseFloat( atk.att.speed )):  10;

					var p:Fast = atk.node.particle;
					a.speed    = p.has.speed? Std.parseFloat( p.att.speed ): 1;
					a.start    = p.has.start? Std.parseFloat( p.att.start ): 0;
					a.image    = p.has.file? p.att.file: "hit.png";
					a.rotated  = p.has.aligned? p.att.aligned=="true": false;
					a.attached = p.has.attached? p.att.attached=="true": false;


					a.damage = [];
					for(dmg in atk.nodes.damage) {
						if(!dmg.has.type) continue;
						var d: Damage = new Damage();
						d.amount   = dmg.has.amount? Std.parseFloat( dmg.att.amount ): 0;
						d.duration = dmg.has.duration? Std.parseFloat( dmg.att.duration ): 0;
						d.type     = dmg.att.type;
						a.damage.push( d );
					}

					m.attacks.push( a );
				}

				// Add to list
				monsterList.set( m.name, m );
			}
		}
		catch(e:String) {
			trace("Monster error " + e);
		}
	}
	public static function create(type:String, x:Float, y:Float) : Monster {
		var def: MonsterDef = monsterList.get(type);
		if(def!=null) return new Monster(def, x, y);
		trace("Monster " + type + " does not exist");
		return null;
	}

	// ====================================================================== //
	public var definition: MonsterDef;
	public var state:   MonsterState; 	// idle, attack, flee
	public var target:  Entity;			// Attack/flee target
	public var closest: Entity;			// Closest entity
	public var health:  Float;			// current health
	public var stateTime: Int;			// time in current state
	public var pointX: Float;			// Additional Info
	public var pointY: Float;			// Additional Info
	public var cooldown: Int;			// attack cooldown
	public var uid: Int;				// UID for debugging
	private static var uuid: Int = 0;

	public function new(def: MonsterDef, x:Float, y:Float) {
		super(x,y);
		definition = def;

		if(def.image!=null) {
			var s:Stamp = new Stamp("graphics/" + def.image);
			s.x = -s.width * 0.5;
			s.y = -s.height * 0.8;
			graphic = s;
			setHitboxTo(graphic);
		}

		type = "solid";
		setState(IDLE);
		health = def.health;
		uid = ++uuid;
	}
	public override function toString() {
		return super.toString() + uid;
	}


	public override function update() {
		if(HXP.distanceSquared(x, y, cast(world, MainScene).player.x, cast(world, MainScene).player.y)>100000) return;
		var see:Bool;
		if(cooldown>0) --cooldown;


		// Do AI
		switch(state) {
		case SLEEP:

		case IDLE:
			if(--stateTime<0) setState(WANDER);

			var r:Int = sense(definition.sense);
			if(r==2) setState(ATTACK, target);
			else if(r==3) setState(FLEE, target);

		case WANDER:
			if(--stateTime<0) setState( HXP.random < 0.7? IDLE: WANDER );
			var lx:Float = x, ly:Float=y;
			moveBy(pointX*definition.speed, pointY*definition.speed, "solid");
			if(lx==x && ly==y) setState(IDLE); // Walked into something

			var r:Int = sense(definition.sense);
			if(r==2) setState(ATTACK, target);
			else if(r==3) setState(FLEE, target);

		case ATTACK:
			see = canSee(target);
			if(see) { pointX = target.x; pointY = target.y; }
			var d:Float = distanceToPoint(pointX, pointY);
			HXP.point.setTo(pointX-x, pointY-y);
			HXP.point.normalize(1.0);

			if(see && d < definition.range) HXP.point.setTo( -HXP.point.x, -HXP.point.y );


			// Spread out from allies
			if(closest!=null) {
				HXP.point2.setTo(closest.x-x, closest.y-y);
				HXP.point2.normalize(0.3);
				HXP.point.add(HXP.point2);
				HXP.point.normalize(1.0);
			}

			moveBy( HXP.point.x * definition.speed, HXP.point.y*definition.speed, "solid" );

			if(!see && d < 10) setState(IDLE); // lost target
			if(!see || --stateTime<=0) {
				var r:Int = sense(definition.sense);
				if(r==2) setState(ATTACK, target);
				else if(r==3) setState(FLEE, target);
				stateTime = 80; // dont switch targets too often
				return;
			}

			// Attack
			if(see && cooldown==0) attack(pointX, pointY);
			if(see && cast(target, Monster).health<=0) setState(IDLE);


		case FLEE:
			see = canSee(target);
			if(see) { pointX = target.x; pointY = target.y;  stateTime=100; }
			else if(--stateTime<0) setState(IDLE);
			// Move
			var lx:Float = x, ly:Float=y;
			moveTowards(pointX, pointY, -definition.speed, "solid");
			if(lx==x && ly==y && see && distanceFrom(target)<32) {
				setState(ATTACK, target); // desperate_attack
				stateTime = 40;
			}

			// Lash out
			if(see && cooldown==0 && distanceFrom(target, true) < 4) {
				attack(pointX, pointY);
			if(see && cast(target, Monster).health<=0) setState(IDLE);
			}

		}

	}

	public function setState(s:MonsterState, tgt:Entity=null) {
		//trace("SetState " + toString() + " " + state + " -> " + s);
		switch(s) {
		case SLEEP:
			pointX = pointY = 0;
			stateTime = 0;
			if(sense(20)>0) setState(ATTACK, target); // Nobody likes waking up
		case IDLE:
			pointX = pointY = 0;
			stateTime = HXP.rand(100);

		case WANDER:
			setNormalisedPoint( HXP.random - 0.5, HXP.random - 0.5 );
			stateTime = HXP.rand(50) + 10;

		case ATTACK:
			target = tgt;
			pointX = tgt.x;
			pointY = tgt.y;
		case FLEE:
			target = tgt;
			pointX = tgt.x;
			pointY = tgt.y;
		}
		state = s;
	}

	private function sense(range:Float) : Int {
		var a:Array<Entity> = [];
		world.collideCircleInto("solid", x, y, range, a);
		var close: Float = range*range;
		var danger: Float = -100;
		var count:Int = 0;
		target = closest = null;
		// analyse
		for(e in a) {
			if(Std.is(e, Monster) && e!=this && cast(e, Monster).health>0) {
				var d2: Float = HXP.distanceSquared(x,y,e.x,e.y);
				var t:Float = cast(e, Monster).definition.threat;
				if(t>danger || (t==danger && d2<close)) { target=e; danger=t; close=d2; }
				++count;
			}
		}
		// Next closest monster
		close = range*range;
		for(e in a) if(Std.is(e, Monster) && e!=this && cast(e,Monster).health>0) {
			var d2: Float = HXP.distanceSquared(x,y,e.x,e.y);
			if(d2 < close && target!=e) { closest = e; close = d2; }
		}


		//danger += definition.aggression;
		danger += 10 - 10 * health / definition.health;
		if(count==0) return 0;		// nobody around
		if(danger < definition.threat) return 1;	// not interested
		if(danger < definition.threat + definition.aggression) return 2;	// attack
		else return 3;	// flee
	}

	public function canSee(t:Entity) : Bool {
		if(t==null) return false;
		collidable = false;
		var e:Entity = world.collideLine("solid", Std.int(x), Std.int(y), Std.int(t.x), Std.int(t.y), 8);
		collidable = true;
		return e == t;
	}

	private function setNormalisedPoint(dx:Float, dy:Float) {
		pointX = dx;
		pointY = dy;
		var l:Float = HXP.distance(pointX, pointY);
		if(l==0) pointY = l = 1;
		pointX/=l;
		pointY/=l;
	}

	// ==================================================================== //

	public function attack(x:Float, y:Float): Bool {
		// select attack
		var t:Int = 0;
		var d: Float = distanceToPoint(x,y, true);
		for(a in definition.attacks) if(isAttackValid(a, d)) t += a.weight;
		if(t==0) return false;

		var r:Int = HXP.rand( t );
		for(a in definition.attacks) {
			if(isAttackValid(a,d) && r<a.weight) {
				world.add( new AttackEffect(this, a, x, y) );
				cooldown = a.cooldown;
				return true;
			} else r -= a.weight;
		}
		return false; // Shouldnt get here
	}
	private function isAttackValid(a:Attack, dist:Float) {
		return a.range >= dist;
	}
	
	public function damage(me:Entity, damage:Damage) {
		if(health<=0) return;

		// Apply damage
		health -= damage.amount;
		if(health<=0) {
			world.remove(this);
			if(definition.name != "snake") {
				cast(world, MainScene).killsText.text = "Kills: " + (cast(world, MainScene).kills++);
			}
		}

		if(definition.name == "player") {
			cast(world, MainScene).healthText.text = "Health: " + health;
		}



		// Retaliate
		setState(ATTACK, me);
	}

}

